@ raspisammlung.s
@ Sammlung von Unterprogrammen in Assembler fuer den Raspberry-Pi

/*
@ ---------------------------------------------------------------------
@ Startsequenz:
@ nur fuer "bare-metal":
        .section .init
        .text
        .global _start
_start:
        mov sp, #0x8000		@ Stack Pointer setzen
        mov r7, #0x20000000
        add r7, #0x200000	@ GPIO-Adresse im r7 behalten
	mov r8, #0x20000000
	add r8, #0x3000		@ Timer-Basisadresse in r8 behalten

@ Variante fuer Auswahl zwischen Linux und "bare-metal":
@.equ LINUX_VERSION, 1  @ auskommentieren fuer "bare-metal"
.include "anfang.s"
*/

@ ---------------------------------------------------------------------
@ GPIO-Unterprogramme:

gpio_init1:	@ einzelner GPIO-Pin initialisieren
@ Parameter: r0=GPIO-Nummer, r1=Ein/Ausgang (0=Eingang, 1=Ausgang)
@            r7=GPIO-Basisadr
	push {r0-r3,lr}
	mov r2, #-4      @ r2=Offset
ini1:	add r2, #4       @ Offset erhoehen
	subs r0, #10	 @ 10 subtrahieren, r0 groesser oder gleich 0? 
	bcs ini1         @ ja--> subtrahieren bis kleiner 0
	add r0, #10      @ zuviel subtrahiertes wieder addieren
	add r3, r0, r0
	add r0, r3	 @ r0 mit 3 multiplizieren
	mov r3, #7
	lsl r3, r0	 @ Maske setzen
	lsl r1, r0       @ zu setzender Wert zurechtschieben
	ldr r0, [r7,r2]  @ schon gesetzte GPIO-Pins lesen
	bic r0, r3       @ Bits gemaess Maske loeschen, alle andern unveraendert
	orr r0, r1       @ gewuenscher Wert setzen
	str r0, [r7,r2]  @ und speichern
	pop {r0-r3,pc}
	
gpio_ein_aus:  @ Parameter: r0=Nummer, r1.bit0=Spannung ein/aus (0=aus, 1=ein),
	       @            r7=GPIO-Basisadresse
        push {r2,lr}        @ Register retten
        mov r2, #1
        lsl r2, r0          @ r2 = (1<<r0)
        tst r1, #1          @ ist erstes Bit in r1 gleich Null?
        streq r2, [r7,#40]  @ ja: Spannung aus
        strne r2, [r7,#28]  @ nein: Spannung ein
        pop {r2,pc}         @ gerettete Register zurueckholen, und Ruecksprung

@ ---------------------------------------------------------------------
@ Wartefunktionen:

microwait:  @ Aufrufparameter: r0 = soviele Microsekunden warten,
	    @                  r8 = Timerbasisadresse
        push {r4-r8,lr}  @ Register retten
	add r8, #4
        ldm r8, {r4,r5}  @ Zaehler des Timers einlesen, 64-Bit
        adds r4, r0
        adc r5, #0	 @ Endzeitpunkt berechnen
micw1:  ldm r8, {r6,r7}  @ Zaehler einlesen
	subs r6, r4
        sbcs r7, r5      @ Endzeitpunkt vom Timerwert subtrahieren
        bcc micw1        @ Springe wenn Endzeitpunkt noch groesser ist
        pop {r4-r8,pc}   @ grettete Register zurueckholen, und Ruecksprung

milliwait:  @ soviele Millisekunden warten wie in r0 steht
	push {r0-r1,lr}
	mov r1, #1000
	mul r0, r1
	bl microwait
	pop  {r0-r1,pc}

@ ---------------------------------------------------------------------
@ Unterprogramme fuer Grafik mit GPU:

MailboxBase:
        .word 0x2000B880    @ Mailbox-Basis-Adresse fuer Kommunikation mit GPU

MailboxWrite:
@ Eingabeparameter: r0=zu sendende Daten, r1=channel (MailboxNummer)
@ Rueckgabe: r0-r2 veraendert
	tst r0,#0xF
	movne pc,lr	@ Ruecksprung wenn Daten falsch
	cmp r1,#15
	movhi pc,lr	@ Ruecksprung wenn channel groesser 15
	push {lr}
	add r1, r0	@ r1 = Daten+channel
	ldr r0, MailboxBase
mbox1:	ldr r2, [r0,#0x18]	@ von Status einlesen
	tst r2, #0x80000000	@ hoechstes Bit gesetzt?
	bne mbox1		@ ja--> warten
	str r1, [r0,#0x20]	@ in Write speichern
	pop {pc}

MailboxRead:
@ Eingabeparameter: r0=channel (MailboxNummer)
@ Rueckgabe: r0=Rueckgabewert (obere 28 Bits), r1-r3 veraendert
	cmp r0,#15
	movhi pc,lr	@ Ruecksprung wenn MailboxNummer groesser 15
	push {r4,lr}
	ldr r4, MailboxBase
mbox2:	ldr r3, [r4,#0x18]	@ von Status einlesen
	tst r3, #0x40000000	@ Bit 30 gesetzt?
	bne mbox2		@ ja--> warten
	ldr r2, [r4,#0]		@ von Read einlesen
	and r1, r2, #0xF	@ r1 = channel vom eingelesenen
	cmp r1, r0		@ stimmt channel ueberein?
	bne mbox2		@ nein--> nochmals warten
	and r0, r2, #0xFFFFFFF0	@ Rueckgabewert ohne channel in r0
	pop {r4,pc}

.section .data
@.align 4   @ unterste 4 Bits muessen 0 sein
.balign 16  @ Adresse muss durch 16 teilbar sein, also unterste 4 Bits 0
.globl FrameBufferInfo
FrameBufferInfo:
	.int 1920 /* #0 Physical Width */
	.int 1080 /* #4 Physical Height */
	.int 1920 /* #8 Virtual Width */
	.int 1080 /* #12 Virtual Height */
	.int 0  /* #16 GPU - Pitch, Anzahl Bytes pro Zeile */
	.int 24 /* #20 Bit Depth */
	.int 0 /* #24 X */
	.int 0 /* #28 Y */
	.int 0 /* #32 GPU - Pointer */
	.int 0 /* #36 GPU - Size */
.section .text
adrFrameBufferInfo:	.word	FrameBufferInfo

.globl InitialiseFrameBuffer
InitialiseFrameBuffer:	@ Eingabeparameter: r0,r1,r2 fuer Breite,Hoehe,Tiefe
	cmp r0, #4096		@ check max width
	cmpls r1,#4096		@ check max height 
	cmpls r2,#32		@ check max bitDepth 
	movhi r0,#0
	movhi pc,lr		@ Ruecksprung bei fehlerhaften Parametern
	push {r1-r4,lr}
	ldr r4, adrFrameBufferInfo  @ r4=Adresse von FrameBufferInfo
	str r0, [r4,#0]
	str r1, [r4,#4]
	str r0, [r4,#8]
	str r1, [r4,#12]
	str r2, [r4,#20]
	b InitF1

.globl InitFrameBuffer
InitFrameBuffer:	@ einfache Variante: voreingestellte Werte benutzen
	push {r1-r4,lr}
	ldr r4, adrFrameBufferInfo  @ r4=Adresse von FrameBufferInfo
InitF1:	orr r0, r4, #0x40000000	@ Anfrage: Initialisieren
	mov r1, #1		@ channel=1
	bl MailboxWrite		@ Anfrage an GPU machen
	mov r0, #1		@ channel=1
	bl MailboxRead		@ sollte auf obige Anfrage 0 zurueck geben
	tst r0,r0		@ Erfolgreich?
	movne r0, #0		@ nein: Rueckgabewert=0
	moveq r0, r4		@ ja: Adresse von FrameBufferInfo
	pop {r1-r4,pc}

clearscreen:	@ Parameter: r0=Farbe
	push {r1-r7,lr}
	ldr r4, adrFrameBufferInfo  @ r4=Adresse von FrameBufferInfo
	ldr r6, [r4,#4]		@ r6=Height
	ldr r7, [r4,#32]	@ r7=Pointer
	and r1, r0, #0xFF	@ r1=R von RGB
	mov r2, r0,lsr#8	@ r2=G von RGB
	mov r3, r2,lsr#8	@ r3=B von RGB
clrs1:	ldr r5, [r4,#0]		@ r5=Widh
clrs2:	strb r1, [r7],#1
	strb r2, [r7],#1
	strb r3, [r7],#1	@ RGB speichern
	subs r5, #1		@ naechster Pixel
	bne clrs2
	subs r6, #1		@ naechste Zeile
	bne clrs1
	pop {r1-r7,pc}

@ ---------------------------------------------------------------------
@ Punkte setzen und lesen:

punkt_setzen:	@ Parameter: r0=x-Wert, r1=y-Wert, r2=Farbe
	push {r2-r4,lr}
	ldr r4, adrFrameBufferInfo  @ r4=Adresse von FrameBufferInfo
	ldr r3, [r4,#16]	@ r3=Pitch
	ldr r4, [r4,#32]	@ r4=Pointer
	mul r3, r1		@ r3=Pitch*y
	add r4, r3
	add r3, r0, r0,lsl#1	@ r3=3*x
	add r4, r3
	strb r2, [r4],#1
	lsr r2,#8
	strb r2, [r4],#1
	lsr r2,#8
	strb r2, [r4]
	pop {r2-r4,pc}

@ noch testen:
punkt_lesen:	@ Parameter: r0=x-Wert, r1=y-Wert
	        @ Rueckgabewert: r0=Farbe
	push {r2-r4,lr}
	ldr r4, adrFrameBufferInfo  @ r4=Adresse von FrameBufferInfo
	ldr r3, [r4,#16]	@ r3=Pitch
	ldr r4, [r4,#32]	@ r4=Pointer
	mul r3, r1		@ r3=Pitch*y
	add r4, r3
	add r3, r0, r0,lsl#1	@ r3=3*x
	add r4, r3
	ldrb r3, [r4],#1
	ldrb r2, [r4],#1
	ldrb r0, [r4]
	lsl r0,#8
	add r0, r2
	lsl r0,#8
	add r0, r3
	pop {r2-r4,pc}

@ noch testen:
	.equ REVERSEBIT, 24	@ Bit fuer Farbumkehrung
	.equ TRANSPARENTBIT, 25	@ Bit fuer Durchsichtigkeit
punkt_loeschen:	@ Parameter: r0=x-Wert, r1=y-Wert, r6=Hintergrundfarbe
	cmp r6, #(1<<REVERSEBIT)	@ Farbe auf Umkehrung gesetzt?
	beq punkt_reverse
	tst r6, #0xFF000000	@ gueltiger RGB-Wert?
	movne pc,lr		@ nein--> Ruecksprung
	push {r2,lr}
	mov r2, r6
	bl punkt_setzen
	pop {r2,pc}

@ noch testen:
punkt_reverse:	@ Farbe eines Pixels umkehren, Parameter: r0=x-Wert, r1=y-Wert
	push {r2-r3,lr}
	mov r3, r0
	bl punkt_lesen
	eor r2, r0, #0xFF
	eor r2, #0xFF00
	eor r2, #0xFF0000
	mov r0, r3
	bl punkt_setzen
	pop {r2-r3,pc}

@ ---------------------------------------------------------------------
@ Linien zeichnen:

waagrechte_linie_zeichnen:
	@ Parameter: r0=x1, r1=y-Wert, r2=Farbe, r3=x2
	push {r0,r4,lr}
	cmp r3, r0
	moveq r4, #0	@ wenn r3==r0 dann r4=0
	movhi r4, #1	@ wenn r3>r0 dann r4= 1
	mvncc r4, #0	@ wenn r3<r0 dann r4= -1
linz1:	bl punkt_setzen
	add r0, r4
	cmp r0, r3
	bne linz1
	pop {r0,r4,pc}

@ (noch optimierbar)
senkrechte_linie_zeichnen:
	@ Parameter: r0=x-Wert, r1=y1, r2=Farbe, r4=y2
	push {r1,r3,lr}
	cmp r4, r1
	moveq r3, #0	@ wenn r4==r1 dann r3=0
	movhi r3, #1	@ wenn r4>r1 dann r3= 1
	mvncc r3, #0	@ wenn r4<r1 dann r3= -1
linz2:	bl punkt_setzen
	add r1, r3
	cmp r1, r4
	bne linz2
	pop {r1,r3,pc}

@ noch ausfuehrlicher testen: (noch optimierbar?)
linie_zeichnen:
	@ Parameter: r0=x1, r1=y1, r2=Farbe, r3=x2, r4=y2
	cmp r0, r3
	beq senkrechte_linie_zeichnen
	cmp r1, r4
	beq waagrechte_linie_zeichnen
	push {r0-r8,lr}
	ldr r5, adrFrameBufferInfo  @ Adresse von FrameBufferInfo
	ldr r6, [r5,#0]
	sub r6, #1	@ Maximaler x-Wert
	cmp r0, r6
	movhi r0, r6	@ Wenn grosser dann auf maximalen Wert setzen
	cmp r3, r6
	movhi r3, r6	@ Wenn grosser dann auf maximalen Wert setzen
	ldr r6, [r5,#4]
	sub r6, #1	@ Maximaler y-Wert
	cmp r1, r6
	movhi r1, r6	@ Wenn grosser dann auf maximalen Wert setzen
	cmp r4, r6
	movhi r4, r6	@ Wenn grosser dann auf maximalen Wert setzen
	bl punkt_setzen
	subs r5, r0, r3		@ dx=x2-x1
	subcc r5, r3, r0	@ oder x1-x2 falls sonst negativ
	subs r6, r1, r4
	subcc r6, r4, r1	@ dy=abs(y2-y1)
	cmp r5, r6		@ ist dx>=dy ?
	bcs linz_xgroesser	@ ja-->
linz_ygroesser:
	mov r8, r6,lsr#1	@ fehler=dy/2
	mov r7, #1		@ xstep=1
	cmp r0, r3		@ x1>x2?
	mvnhi r7, #0		@ ja: xstep = -1
linz3:	add r1, #1		@ y=y+1
	subs r8, r5		@ fehler=fehler-dx
	addcc r0, r7		@ wenn fehler<0 dann x1+=xstep
	addcc r8, r6		@               und fehler+=dy
	bl punkt_setzen
	cmp r1, r4		@ y==y2?
	bne linz3		@ nein-->
	pop {r0-r8,pc}
linz_xgroesser:
	mov r8, r5,lsr#1	@ fehler=dx/2
	mov r7, #1		@ ystep=1
	cmp r1, r4		@ y1>y2?
	mvnhi r7, #0		@ ja: ystep = -1
linz4:	add r0, #1		@ x=x+1
	subs r8, r6		@ fehler=fehler-dy
	addcc r1, r7		@ wenn fehler<0 dann y1+=ystep
	addcc r8, r5		@               und fehler+=dx
	bl punkt_setzen
	cmp r0, r3		@ x==x2?
	bne linz4		@ nein-->
	pop {r0-r8,pc}

/* provi. Kreise und Ellipse auskommentiert
@ ---------------------------------------------------------------------
@ Kreise zeichnen:
@ Bresenham-Algorithus fuer Kreis
kreis_zeichnen:	@ Parameter: r0=x, r1=y-Wert von Mitte, r2=Farbe, r3=Radius
	push {r0,r1,r4-r8,lr}
	mov r4, r0	@ r4=xmitte
	mov r5, r1	@ r5=ymitte
	mov r6, r3	@ fehler=r
	mov r7, r3	@ x=r
	mov r8, #0	@ y=0
	add r0, r4, r7	@ xmitte+x, (r1=ymitte+y ist schon richtig)
	bl punkt_setzen
	subs r0, r4, r7	@ xmitte-x
	blcs punkt_setzen
	mov r0, r4
	add r1, r5, r3	@ ymitte+r
	bl punkt_setzen
	subs r1, r5, r3	@ ymitte-r
	blcs punkt_setzen
kr1:	add r1, r8, r8
	add r1, #1	@ dy=2*y+1
	add r8, #1	@ y++
	subs r6, r1	@ fehler -= dy
	bcs kr2		@ wenn fehler>=0 -->
	add r6, r7,lsl#1	@ sonst: fehler += 2*x-1
	sub r6, #1
	sub r7, #1		@ und x erniedrigen
kr2:	add r0, r4, r7	@ xmitte+x
	add r1, r5, r8	@ ymitte+y
	bl punkt_setzen	@ 1. Oktant
	subs r0, r4, r7	@ xmitte-x
	bcc kr3		@ wenn kleiner 0 dann 2.u.3. Oktant ueberspringen
	bl punkt_setzen	  @ 2. Oktant
	subs r1, r5, r8	  @ ymitte-y
	blcs punkt_setzen @ 3. Oktant (sofern ymitte-y nicht kleiner 0)
kr3:	add r0, r4, r7	@ xmitte+x
	subs r1, r5, r8	@ ymitte-y
	blcs punkt_setzen @ 4. Oktant
	add r0, r4, r8	@ xmitte+y
	add r1, r5, r7	@ ymitte+x
	bl punkt_setzen	@ 5. Oktant
	subs r0, r4, r8	@ xmitte-y
	bcc kr4		@ wenn kleiner 0 dann 6.u.7. Oktant ueberspringen
	bl punkt_setzen		@ 6. Oktant
	subs r1, r5, r7		@ ymitte-x
	blcs punkt_setzen	@ 7. Oktant
kr4:	add r0, r4, r8	  @ xmitte+y
	subs r1, r5, r7	  @ ymitte-x
	blcs punkt_setzen @ 8. Oktant
	cmp r8, r7	@ y<x ?
	bcc kr1
	pop {r0,r1,r4-r8,pc}

@ ---------------------------------------------------------------------
@ Bresenham-Algorithus fuer Ellipse
ellipse_zeichnen:  @ Parameter: r0=x, r1=y-Wert von Mitte, r2=Farbe, r3=a=Radiusx, r4=b=Radiusy
	push {r0,r1,r5-r11,lr}
	mov r5, r0	@ int r5=xmitte
	mov r6, r1	@ int r6=ymitte
	mov r10, r3
	mul r10, r3	@ long r10=a*a
	mov r11, r4
	mul r11, r4	@ long r11=b*b
	mov r0, r4,lsl#1 @ 2*b
	sub r0, #1	 @ 2*b-1
	mul r0, r10	 @ (2*b-1)*a*a
	mov r9, r11	@ b*b
	sub r9, r0	@ long r9=err=b*b-(2*b-1)*a*a
	mov r7, #0	@ int r7=dx=0
	mov r8, r4	@ int r8=dy=b
el1:	add r0, r5, r7	@ xmitte+dx
	add r1, r6, r8	@ ymitte+dy
	bl punkt_setzen	@ 1.Quadrant
	subs r0, r5, r7	@ xmitte-dx
	bcc el2		@ wenn negativ, dann 2.u.3. Quadrant auslassen
	bl punkt_setzen		@ 2.Quadrant
	subs r1, r6, r8		@ ymitte-dy
	blcs punkt_setzen	@ 3.Quadrant
el2:	add r0, r5, r7		@ xmitte+dx
	subs r1, r6, r8		@ ymitte-dy
	blcs punkt_setzen	@ 4.Quadrant
	mov r1, r9,lsl#1	@ r1=e2=2*err
	mov r0, r7,lsl#1	@ 2*dx
	add r0, #1		@ 2*dx+1
	mul r0, r11		@ r0=(2*dx+1)*b*b
	cmp r1, r0		@ e2<r0 ?
	addlt r7, #1		@ ja: dx++
	addlt r9, r0		@ ja: err += r0
	mov r0, r8,lsl#1	@ 2*dy
	sub r0, #1		@ 2*dy-1
	mul r0, r10		@ (2*dy-1)*a*a
	rsb r0, #0		@ r0 = -(2*dy-1)*a*a
	cmp r0, r1		@ e2>r0 ? (Vorzeichenbehaftet!)
	bge el1			@ nein -->
	add r9, r0		@ ja: err += r0
	subs r8, #1		@ ja: dy--
	bcs el1			@ wenn dy>=0 -->
el3:	cmp r7, r3	@ dx<a ?
	bcs el4		@ nein -->
	add r7, #1	@ dx++
	add r0, r5, r7	@ xmitte+dx
	mov r1, r6	@ ymitte
	bl punkt_setzen	@ Spitze der Ellipso vollenden
	subs r0, r5, r7	@ xmitte-dx
	blcs punkt_setzen
	b el3
el4:	pop {r0,r1,r5-r11,pc}

provi. Kreise und Ellipse auskommentiert */
	
@ ---------------------------------------------------------------------
@ Text zeichnen:

.include "fonttabellen.s"
font_zeichnen:	@ Parameter: r0,r1 = x,y-Position links unten des Zeichens, r2=Farbe,
		@   r3=Zeichencode, r4=Breite*256+Hoehe, r5=Fonttyp*256+Skalierfaktor, r6=Hintergrundfarbe
	push {r0-r1,r3-r11,lr}
	ands r8, r5, #0xFF		@ r8=Skalierfaktor
	moveq r8, #1 			@ mindestens 1
	lsrs r5, #8			@ r5=Fonttyp
	moveq r5, #1			@ mindestens 1
	cmp r5, #1			@ Fonttyp==1 ?
	cmpeq r4, #((12<<8)+16)		@ und Groesse 12x16 ?
	beq font_12x16_typ1
	@ hier weiter Abfragen fuer andere Fonts
font_unbekannt:				@ bei unbekanntem Font leeres Viereck zeichnen
	mov r5, r4,lsr#8		@ r5=Breite
	mul r5, r8			@ r5=Breite*Skalierfaktor
	and r6, r4, #0xFF		@ r6=Hoehe
	mul r6, r8			@ r6=Hoehe*Skalierfaktor
	add r3, r0, r5
	bl waagrechte_linie_zeichnen	@ Linie von x,y nach x2,y
	mov r3, r0
	subs r4, r1, r6			@ r4=y2=y-Hoehe
	movcc r4, #0			@ bei negativem Wert auf 0 setzen
	bl senkrechte_linie_zeichnen	@ Linie von x,y nach x,y2
	add r3, r0, r5
	mov r7, r1			@ y1 retten
	mov r1, r4			@ y1 auf y2 setzen
	bl waagrechte_linie_zeichnen	@ Linie von x,y2 nach x2,y2
	mov r1, r7			@ y1 zurueckholen
	mov r0, r3
	bl senkrechte_linie_zeichnen	@ Linie von x2,y nach x2,y2
	b font_ret
font_12x16_typ1:	@ Typ1: Bits von links nach rechts, Bytes von oben nach unten
	ldr r4, adr_fonttabelle_12x16_typ1
	add r4, r3,lsl#5	@ r4=Zeiger auf Daten des aktuellen Zeichens (r3*32 weil 32 Bytes pro Zeichen)
	subs r1, r8,lsl#4	@ y von unten links nach oben links setzen (16*Skalierfaktor subtrahieren)
	movcc r1, #0		@ Bei Fehler y=0 setzen
	mov r11, r0		@ Anfangswert von x
	mov r5, #16		@ 16 Bytes behandeln
font1:	mov r9, r8
font1s:	ldrb r3, [r4,#0]	@ Byte lesen
	ldrb r7, [r4,#1]	@ zweites Byte lesen
	lsl r3, #8
	add r3, r7
	mov r7, #12		@ 12 Bits behandeln
font2:	mov r10, r8
font2s:	tst r3, #0x8000		@ Bit gesetzt?
	blne punkt_setzen
	bleq punkt_loeschen
	add r0, #1		@ x erhoehen
	subs r10, #1		@ mehrmals bei Skalierfaktor>1
	bne font2s
	lsl r3, #1		@ naechstes Bit
	subs r7, #1
	bne font2
	mov r0, r11		@ x wieder auf Anfangswert
	add r1, #1		@ y erhoehen
	subs r9, #1		@ mehrmals bei Skalierfaktor>1
	bne font1s
	add r4, #2
	subs r5, #1		@ naechstes Byte
	bne font1
font_ret:
	pop {r0-r1,r3-r11,pc}

.globl TextParameter
TextParameter:
	.int 12 /* #0 Zeichenbreite */
	.int 16 /* #4 Zeichenhoehe */
	.int 0  /* #8 Zeichenabstand */
	.int 4  /* #12 Zeilenabstand */
	.int 1  /* #16 Skalierfaktor */
	.int 0xFFFFFF /* #20 Textfarbe */
	.int 0x000000 /* #24 Hintergrundfarbe */
.section .text
adrTextParameter: .word TextParameter
/*
Beispiele fuer Farbe:
	mov r2, #0		@ schwarz
	mov r2, #0xFF		@ rot
	mov r2, #0xFF00		@ gruen
	mov r2, #0xFF0000	@ blau
	mvn r2, #0xFF000000	@ weiss  (mvn = Bits umgekehrt!)
*/
setfontsize:	@ r0=Zeichenbreite, r1=Zeichenhoehe, r2=Zeichenabstand, r3=Zeilenabstand
	push {r8,lr}
	ldr r8, adrTextParameter
	str r0, [r8],#4
	str r1, [r8],#4
	str r2, [r8],#4
	str r3, [r8],#4
	pop {r8,pc}

setfontscale:	@ r0=Skalierfaktor
	push {r8,lr}
	ldr r8, adrTextParameter
	str r0, [r8,#16]	@ Skalierfaktor
	pop {r8,pc}

setfontcolors:	@ r2=Vordergrundfarbe, r6=Hintergrundfarbe
	push {r8,lr}
	ldr r8, adrTextParameter
	str r2, [r8,#20]	@ Textfarbe
	str r6, [r8,#24]	@ Hintergrundfarbe
	pop {r8,pc}

printchar:	@ Parameter: r0,r1 = x,y-Position des Zeichens, r3=Zeichen
	        @ Rueckgabe: r0,r1 = neue Position
		@ Text zeichnen mit einem 12x16 Typ1-Font
	push {r2,r4-r7,lr}
	ldr r7, adrTextParameter
	ldr r5, [r7,#0]		@ Breite
	ldr r4, [r7,#4]		@ Hoehe
	add r4, r5,lsl#8	@ r4=Breite*256+Hoehe
	ldr r5, [r7,#16]	@ Skalierfaktor
	add r5, #(1<<8)		@ Fonttyp=1
	ldr r2, [r7,#20]	@ Vordergrundfarbe laden 
	ldr r6, [r7,#24]	@ Hintergrundfarbe laden
	bl font_zeichnen
	ldr r2, [r7,#16]	@ r2=Skalierfaktor
	ldr r5, [r7,#0]		@ r5=Zeichenbreite
	mul r5, r2
	ldr r6, [r7,#8]		@ r6=Zeichenabstand
	mul r6, r2
	add r0, r5
	add r0, r6		@ (Zeichenbreite+Zeichenabstand)*Skalierfaktor zu x addieren
	ldr r4, adrFrameBufferInfo
	ldr r6, [r4,#0]		@ Bildbreite einlesen
	sub r6, r5		@ r6=Bildbreite-Zeichenbreite*Skalierfaktor
	ldr r5, [r7,#4]		@ r5=Zeichenhoehe
	mul r5, r2
	cmp r0, r6		@ noch Platz fuer ein Zeichen?  r6>r0 ?
	bcc prch1		@ ja -->
	mov r0, #0		@ nein: eine neue Zeile beginnen
	ldr r6, [r7,#12]	@ Zeilenabstand
	mul r6, r2
	add r1, r5
	add r1, r6		@ und (Buchstabenhoehe+Zeilenabstand)*Skalierfaktor zu y addieren
prch1:	ldr r6, [r4,#4]		@ Bildhoehe einlesen
	cmp r1, r6
	movcs r1, r5		@ wenn y zu gross, wieder mit erster Zeile beginnen (y=0+Zeichenhoehe)
	pop {r2,r4-r7,pc}

printstring:	@ Parameter: r0,r1 = x,y-Position des ersten Zeichens, r3=String mit 0 abgeschlossen
		@ Rueckgabe: r0,r1 = neue Position
	push {r3-r4,lr}
	mov r4, r3
pr1:	ldrb r3, [r4],#1
	cmp r3,#0
	beq pr2
	bl printchar
	b pr1
pr2:	pop {r3-r4,pc}

printblank:
	push {r3,lr}
	mov r3, #' '
	bl printchar
	pop {r3,pc}

divmod:	@ Positive Zahlen dividieren: r2 /= r3, r4=Rest
	push {r5-r6,lr}
	mov r4, #0	@ Rest=0
	mov r5, #0	@ Resultat=0
	mov r6, #32	@ 32 Bits behandeln
divm1:	lsl r5, #1
	lsl r4, #1
	lsls r2, #1
	addcs r4, #1	@ Bits von r2 nach links ins r4 schieben
	cmp r4, r3	@ r4>=r3
	addcs r5, #1	@ ja: Resultat-Bit setzen
	subcs r4, r3	@ ja: und r3 von r4 Subtrahieren
	subs r6, #1	@ naechstes Bit
	bne divm1
	mov r2, r5	@ r2=Resultat, r4=Rest
	pop {r5-r6,pc}

	.data
Scratch1: .space 128	@ Platz fuer temporaere Daten
Scratch2: .space 128	@ Platz fuer temporaere Daten
	.text
adrScratch1:	.word Scratch1
adrScratch2:	.word Scratch2

printzahl:	@ r0=x, r1=y, r2=Platz, r3=Zahl
		@ Platz ist Formatangabe, 0=unformatiert, >1=Leerstellen links anfuegen
		@ oder wenn hoechstwertiges Byte von r2 ungleich null: Zeichen zum Auffuellen (statt Leerzeichen)
	push {r2-r8,lr}
	mov r8, r2		@ r8=Platz
	ldr r6, adrScratch2
	mov r4, #0
	strb r4, [r6],#-1	@ 0 als Stringabschluss
	and r2, #0xFF
	sub r7, r6, r2		@ r7=Puffergrenze
	ldr r2, adrScratch1
	sub r2, #1
	tst r8, r8		@ Platz==0 ?
	moveq r7, r2		@ ja: Puffergrenze=scratch1-1
	beq prza0		@ ja: -->
	cmp r7, r2		@ r7<r2 ?
	movcc r7, r2		@ ja: r7=Puffergrenze korrigieren
prza0:	mov r2, r3		@ r2 = zu druckende Zahl
	tst r2, #0x80000000	@ Zahl negativ?
	rsbne r2, r2, #0	@ ja: Zahl positiv machen
	movne r5, #'-'		@ ja: Vorzeichen -
	moveq r5, #0		@ nein: kein Vorzeichen
	mov r3, #10
prza1:	bl divmod		@ r2 /= r3, r4=Rest
	add r4, #'0'
	strb r4, [r6],#-1
	cmp r6, r7		@ Puffergrenze erreicht?
	beq prza2		@ ja -->
	tst r2, r2		@ restliche Zahl==0 ?
	bne prza1		@ nein -->
	tst r5, r5
	strneb r5, [r6],#-1	@ Vorzeichen speichern
	tst r8, r8		@ unformatiert?
	beq prza2		@ ja -->
	movs r5, r8,lsr#24
	moveq r5, #' '
prza3:	cmp r6, r7		@ Puffergrenze erreicht?
	beq prza2		@ ja -->
	strb r5, [r6],#-1	@ Mit Leerstellen auffuellen
	b prza3
prza2:	add r3, r6, #1		@ r3=String
	bl printstring
	pop {r2-r8,pc}

printhexzahl:	@ r0=x, r1=y, r2=Platz, r3=Zahl
		@ Platz ist Formatangabe, 0=unformatiert, >1=Leerstellen links anfuegen
		@ oder wenn hoechstwertiges Byte von r2 ungleich null: Zeichen zum Auffuellen (statt Leerzeichen)
	push {r2-r8,lr}
	mov r8, r2		@ r8=Platz
	ldr r6, adrScratch2
	mov r4, #0
	strb r4, [r6],#-1	@ 0 als Stringabschluss
	and r2, #0xFF
	sub r7, r6, r2		@ r7=Puffergrenze
	ldr r2, adrScratch1
	sub r2, #1
	tst r8, r8		@ Platz==0 ?
	moveq r7, r2		@ ja: Puffergrenze=scratch1-1
	beq prhe0		@ ja: -->
	cmp r7, r2		@ r7<r2 ?
	movcc r7, r2		@ ja: r7=Puffergrenze korrigieren
prhe0:	mov r2, r3		@ r2 = zu druckende Zahl
	tst r2, #0x80000000	@ Zahl negativ?
	rsbne r2, r2, #0	@ ja: Zahl positiv machen
	movne r5, #'-'		@ ja: Vorzeichen -
	moveq r5, #0		@ nein: kein Vorzeichen
prhe1:	and r4, r2, #0xF	@ r4=Rest bei Division durch 16
	lsr r2, #4		@ r2 /= 16
	cmp r4, #10
	addcc r4, #'0'
	addcs r4, #'A'-10
	strb r4, [r6],#-1
	cmp r6, r7		@ Puffergrenze erreicht?
	beq prhe2		@ ja -->
	tst r2, r2		@ restliche Zahl==0 ?
	bne prhe1		@ nein -->
	tst r5, r5
	strneb r5, [r6],#-1	@ Vorzeichen speichern
	tst r8, r8		@ unformatiert?
	beq prhe2		@ ja -->
	movs r5, r8,lsr#24
	moveq r5, #' '
prhe3:	cmp r6, r7		@ Puffergrenze erreicht?
	beq prhe2		@ ja -->
	strb r5, [r6],#-1	@ Mit Leerstellen auffuellen
	b prhe3
prhe2:	add r3, r6, #1		@ r3=String
	bl printstring
	pop {r2-r8,pc}

print_newline:
	push {r2-r4,lr}
	ldr r4, adrTextParameter
	ldr r2, [r4,#16]	@ Skalierfaktor 
	ldr r3, [r4,#12]	@ Zeilenabstand
	mul r3, r2
	add r1,	r3		@ y um skalierten Zeilenabstand erhoehen
	ldr r3, [r4,#4]		@ Zeichenhoehe
	mul r3, r2
	add r1,	r3		@ y um skalierte Zeichenhoehe erhoehen
	ldr r4, adrFrameBufferInfo
	ldr r2, [r4,#4]		@ Bildhoehe
	cmp r1, r2		@ unterer Rand erreicht?
	movcs r1, r3		@ ja: wieder oben beginnen
	mov r0, #0		@ x=0
	pop {r2-r4,pc}

@ ---------------------------------------------------------------------
@ Ende der Routinensammlung.
